# Time Profiling Tool
# By Braddock Gaskill, Dec 2012

import logging
log = logging.getLogger(__name__)
#log.addHandler(logging.NullHandler())
log.addHandler(logging.StreamHandler())
log.setLevel(logging.DEBUG)

from time import time
from functools import wraps
import threading
from os import getpid


start_time = time()


global_active = False

try:
    from psutil import Process
    def get_memory_usage():
        pid = getpid()
        proc = Process(pid)
        mem = proc.get_memory_info()
        return mem
except Exception:
    def get_memory_usage():
        return 0


class TimePro(object):
    def __init__(self):
        global global_active
        self.times = {}
        self.ancestors = []
        self.active = global_active

    def activate(self):
        self.reset()
        self.active = True

    def deactivate(self):
        self.active = False
        self.reset()

    def start(self, name):
        if not self.active:
            return
        if len(self.ancestors) > 0:
            fullname = self.ancestors[-1] + "->" + name
        else:
            fullname = name
        self.ancestors.append(name)
        r = self.times.get(fullname, None)
        if r is None:
            r = {
                'count': 0,
                'total': 0.0,
                'max': -1.0,
                'min': -1.0,
                'mem_res_total': 0,
                'mem_virt_total': 0
            }
        r['count'] += 1
        r['start_time'] = time()
        mem = get_memory_usage()
        r['start_mem_res'] = mem.rss
        r['start_mem_virt'] = mem.vms
        self.times[fullname] = r

    def end(self, name):
        if not self.active:
            return
        self.ancestors.pop()
        if len(self.ancestors) > 0:
            fullname = self.ancestors[-1] + "->" + name
        else:
            fullname = name
        r = self.times[fullname]
        t0 = r['start_time']
        dt = time() - t0
        mem = get_memory_usage()
        r['total'] += dt
        if r['count'] == 1:
            r['max'] = dt
            r['min'] = dt
        r['max'] = max(r['max'], dt)
        r['min'] = min(r['min'], dt)
        r['mem_res_total'] += mem.rss - r['start_mem_res']
        r['mem_virt_total'] += mem.vms - r['start_mem_virt']

    def log_all(self):
        if not self.active:
            return
        log.debug("TOTAL RUNTIME " + str(time() - start_time))
        log.debug("COUNT  TOTAL      AVG     MAX     MIN   dMEM(MB) NAME")
        for name in self.times:
            r = self.times[name]
            dmem = float(r['mem_res_total']) / float(r['count']) / float(1024 * 1024)
            log.debug("%5i  %6.3f  %6.5f  %6.5f  %6.5f  %6i  %s"
                      % (r['count'], r['total'], r['total'] / float(r['count']), r['max'], r['min'], int(dmem), name))

    def reset(self):
        self.times = {}


threadlocal = threading.local()


def timepro():
    """Get the timepro singleton for this thread.
    Create a new instance if it doesn't exist for this thread."""
    v = getattr(threadlocal, 'timepro', None)
    if v is None:
        threadlocal.timepro = TimePro()
    return threadlocal.timepro


def start(name):
    """Start a named timer"""
    timepro().start(name)


def end(name):
    """Stop a named timer"""
    timepro().end(name)


def log_all():
    """Log all time profiling information for this thread"""
    timepro().log_all()


def profile():
    """This is a decorator you can wrap functions in to get
    timing information"""
    def wrapper1(fn):
        @wraps(fn)
        def wrapper2(*args, **kwargs):
            timepro().start(fn.func_name)
            r = fn(*args, **kwargs)
            timepro().end(fn.func_name)
            return r
        return wrapper2
    return wrapper1


def profile_and_print():
    """This is a decorator you can wrap functions in to get
    timing information"""
    def wrapper1(fn):
        @wraps(fn)
        def wrapper2(*args, **kwargs):
            timepro().start(fn.func_name)
            r = fn(*args, **kwargs)
            timepro().end(fn.func_name)
            timepro().log_all()
            return r
        return wrapper2
    return wrapper1


def reset():
    """Reset all timers for this thread"""
    timepro().reset()



# REMOVE: OBSOLETE: use zimpy.py instead
# Extract articles from ZIM files using zimdump (for now)
# Must use a patched version of zimdump
# By Braddock Gaskill May 2013

from subprocess import Popen, PIPE
import os
import re

import kiwix
from config import config


def parse_zimdump_line(line):
    line = line.strip()
    g = re.match("([a-z\- ]+): +(.+)", line)
    if g is not None:
        return g.groups()[0], g.groups()[1]
    else:
        return None, None


def parse_zimdump_info(text):
    d = {}
    lines = text.split("\n")
    for line in lines:
        name, value = parse_zimdump_line(line)
        if name is not None:
            d[name] = value
    return d


def get_article_info_by_url(zimfile, namespace, url, cwd='.'):
    name = "X" + namespace + "/" + url
    exe = config().get_path('KIWIX', 'zimdump')
    cmd = [exe,
           '-i', '-u', name,
           zimfile]
    cmd = [x.encode('utf-8') for x in cmd]
    p = Popen(cmd, stdout=PIPE, cwd=cwd)
    data = p.stdout.read()
    if len(data) == 0:
        print "ZERO LENGTH FILE"
        return None
    return parse_zimdump_info(data)


def get_zimfile_info(zimfile, cwd='.'):
    """Return info about the zim file itself"""
    exe = config().get_path('KIWIX', 'zimdump')
    cmd = [exe,
           '-F', '-v',
           zimfile]
    #print string.join(cmd, " ")
    cmd = [x.encode('utf-8') for x in cmd]
    p = Popen(cmd, stdout=PIPE, cwd=cwd)
    data = p.stdout.read()
    info = parse_zimdump_info(data)
    return info


def get_article_info_by_index(zimfile, idx, cwd='.'):
    exe = config().get_path('KIWIX', 'zimdump')
    cmd = [exe,
           '-i', '-o', str(idx),
           zimfile]
    #print string.join(cmd, " ")
    cmd = [x.encode('utf-8') for x in cmd]
    p = Popen(cmd, stdout=PIPE, cwd=cwd)
    data = p.stdout.read()
    if len(data) == 0:
        print "ZERO LENGTH FILE"
        return None
    return parse_zimdump_info(data)


def get_article_data_by_index(zimfile, index, cwd='.'):
    exe = config().get_path('KIWIX', 'zimdump')
    cmd = [exe,
           '-d', '-o', str(index),
           zimfile]
    #print string.join(cmd, " ")
    cmd = [x.encode('utf-8') for x in cmd]
    p = Popen(cmd, stdout=PIPE, cwd=cwd)
    data = p.stdout.read()
    if len(data) == 0:
        print "ZERO LENGTH FILE"
    return data


def get_article_by_url(zimfile, namespace, url, cwd='.'):
    info = get_article_info_by_url(zimfile, namespace, url, cwd=cwd)
    if not bool(int(info.get('redirect'))):
        data = get_article_data_by_index(zimfile, info['idx'], cwd=cwd)
    else:
        data = None
    return data, info


def get_article_by_index(zimfile, idx, cwd='.'):
    info = get_article_info_by_index(zimfile, idx, cwd=cwd)
    if not bool(int(info.get('redirect'))):
        data = get_article_data_by_index(zimfile, info['idx'], cwd=cwd)
    else:
        data = None
    return data, info


def replace_paths(top_url, html):
    replace = u"\\1\\2" + top_url + "/\\3/"
    html = re.sub(u'(href|src)(=["\']/)([A-Z\-])/', replace, html)
    html = re.sub(u'(@import[ ]+)(["\']/)([A-Z\-])/', replace, html)
    return html


class ZimArticle(object):
    def __init__(self, data, info):
        self.data = data
        self.url = info['url']
        self.title = info['title']
        self.idx = int(info['idx'])
        self.namespace = info['namespace']
        self.redirect = bool(int(info['redirect']))
        self.mime_type = info.get('mime-type', None)
        self.size = info.get('article size', None)
        self.redirect_index = info.get('redirect index', None)


class ZimFile(object):
    def __init__(self, book):
        self.book = book

    def get_filename(self):
        kiwix_dir = config().get_path('KIWIX', 'wikipedia_kiwix_dir')
        path = os.path.join(kiwix_dir, self.book['path'])
        return path

    def get_article_by_url(self, namespace, url):
        kiwix_dir = config().get_path('KIWIX', 'wikipedia_kiwix_dir')
        path = self.get_filename()
        data, info = get_article_by_url(path, namespace, url, cwd=kiwix_dir)
        article = ZimArticle(data, info)
        while article.redirect:
            data, info = get_article_by_index(path, article.redirect_index, cwd=kiwix_dir)
            article = ZimArticle(data, info)
        return article

    def get_article_by_index(self, idx):
        kiwix_dir = config().get_path('KIWIX', 'wikipedia_kiwix_dir')
        path = self.get_filename()
        data, info = get_article_by_index(path, idx, cwd=kiwix_dir)
        article = ZimArticle(data, info)
        while article.redirect:
            data, info = get_article_by_index(path, article.redirect_index, cwd=kiwix_dir)
            article = ZimArticle(data, info)
        return article

    def get_info(self):
        kiwix_dir = config().get_path('KIWIX', 'wikipedia_kiwix_dir')
        info = get_zimfile_info(self.get_filename(), cwd=kiwix_dir)
        return info

    def get_main_page(self):
        info = self.get_info()
        return self.get_article_by_index(info.get('main page', 1))


class Library(object):
    def __init__(self, library_filename):
        self.books = kiwix.parse_library(library_filename)
        self.readableToBooks = dict([(x['humanReadableId'], ZimFile(x)) for x in self.books])

    def get_article_by_url(self, humanReadableId, namespace, url):
        zimfile = self.get_zimfile(humanReadableId)
        if zimfile is None:
            return None
        article = zimfile.get_article_by_url(namespace, url)
        return article

    def get_zimfile(self, humanReadableId):
        return self.readableToBooks.get(humanReadableId, None)



# -*- coding: utf-8 -*-

from flask import Markup

from flask.ext.wtf import Form, ValidationError
from flask.ext.wtf import (HiddenField, BooleanField, TextField,
        PasswordField, SubmitField)
from flask.ext.wtf import Required, Length, EqualTo, Email
from flask.ext.wtf.html5 import EmailField

from .utils import (PASSWORD_LEN_MIN, PASSWORD_LEN_MAX, 
        USERNAME_LEN_MIN, USERNAME_LEN_MAX)


class LoginForm(Form):
    next = HiddenField()
    login = TextField(u'Username or email', [Required()])
    password = PasswordField('Password', [Required(), Length(PASSWORD_LEN_MIN, PASSWORD_LEN_MAX)])
    remember = BooleanField('Remember me')
    submit = SubmitField('Sign in')


class SignupForm(Form):
    next = HiddenField()
    email = EmailField(u'Email', [Required(), Email()],
            description=u"What's your email address?")
    password = PasswordField(u'Password', [Required(), Length(PASSWORD_LEN_MIN, PASSWORD_LEN_MAX)],
            description=u'%s characters or more! Be tricky.' % PASSWORD_LEN_MIN)
    name = TextField(u'Choose your username', [Required(), Length(USERNAME_LEN_MIN, USERNAME_LEN_MAX)],
            description=u"Don't worry. you can change it later.")
    agree = BooleanField(u'Agree to the ' + 
        Markup('<a target="blank" href="/terms">Terms of Servic</a>'), [Required()])
    submit = SubmitField('Sign up')

    def validate_name(self, field):
        if User.query.filter_by(name=field.data).first() is not None:
            raise ValidationError(u'This username is taken')

    def validate_email(self, field):
        if User.query.filter_by(email=field.data).first() is not None:
            raise ValidationError(u'This email is taken')


class RecoverPasswordForm(Form):
    email = EmailField(u'Your email', [Email()])
    submit = SubmitField('Send instructions')


class ChangePasswordForm(Form):
    activation_key = HiddenField()
    password = PasswordField(u'Password', [Required()])
    password_again = PasswordField(u'Password again', [EqualTo('password', message="Passwords don't match")]) 
    submit = SubmitField('Save')


class ReauthForm(Form):
    next = HiddenField()
    password = PasswordField(u'Password', [Required(), Length(PASSWORD_LEN_MIN, PASSWORD_LEN_MAX)])
    submit = SubmitField('Reauthenticate')



class EndPointDescription(object):
    """Description of url for use with pagination"""
    def __init__(self, endpoint, values):
        """matches parameters used in url_for(endpoint, values) function
        :param endpoint: string description of endpoint such as blueprint_class.funcname
        :param values: dictionary of parameters to pass to url_for"""
        self._endpoint = endpoint
        self._values = values or {}

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def values(self):
        return self._values



# ZIM file URL views (for Wikipedia)
from flask import Blueprint, Response, render_template

from zim import Library, replace_paths
from config import config

blueprint = Blueprint('zim_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route('/<humanReadableId>')
def zim_main_page_view(humanReadableId):
    """Returns the main page of the zim file"""
    library_xml = config().get_path('KIWIX', 'library')
    lib = Library(library_xml)
    zimfile = lib.get_zimfile(humanReadableId)
    try:
        article, mimetype, ns = zimfile.get_main_page()
        html = mangle_article(article, mimetype, humanReadableId)
        return Response(html, mimetype=mimetype)
    except OSError as e:
        html = "<html><body>"
        html += "<p>Error accessing article.  Possible failure to run zimdump command</p>"
        html += "<p>zimdump = " + config().get_path('KIWIX', 'zimdump') + "</p>\n"
        html += "<p>Exception: " + str(e) + "</p>\n"
        html += "</body></html>"
        return Response(html)


def mangle_article(html, mimetype, humanReadableId):
    if mimetype in ['text/html; charset=utf-8', 'stylesheet/css', 'text/html']:
        try:
            html = html.decode('utf-8')
        except UnicodeDecodeError:
            try:
                print "utf-8 decoding failed, falling back to latin1"
                html = html.decode('latin1')
            except:
                print "utf-8 and latin1 decoding failed"
                return html
        html = replace_paths("iiab/zim/" + humanReadableId, html)
    return html


@blueprint.route('/<humanReadableId>/<namespace>/<path:url>')
def zim_view(humanReadableId, namespace, url):
    library_xml = config().get_path('KIWIX', 'library')
    lib = Library(library_xml)
    article, mimetype, ns = lib.get_article_by_url(humanReadableId, namespace, url)
    html = mangle_article(article, mimetype, humanReadableId)
    return Response(html, mimetype=mimetype)


@blueprint.route('/iframe/<humanReadableId>')
def iframe_main_page_view(humanReadableId):
    url = '/iiab/zim/' + humanReadableId + '/'
    return render_template('zim_iframe.html', url=url)


@blueprint.route('/iframe/<humanReadableId>/<namespace>/<path:url>')
def iframe_view(humanReadableId, namespace, url):
    url = '/iiab/zim/' + humanReadableId + '/' + namespace + '/' + url
    return render_template('zim_iframe.html', url=url)



from flask import Blueprint, render_template

from config import config
from kiwix import parse_library, get_languages


blueprint = Blueprint('wikipedia_views', __name__,
                      template_folder='templates')


@blueprint.route('/')
def wikipedia_view():
    library_xml = config().get_path('KIWIX', 'library')
    kiwix_url = config().get('KIWIX', 'url')
    library = parse_library(library_xml)
    langs = get_languages(library)
    ajax = False
    if kiwix_url == "/iiab/zim":
        kiwix_url = "/iiab/zim/iframe"
        ajax = True
    return render_template('wikipedia_index.html', languages=langs, kiwix_url=kiwix_url, ajax=ajax)



# Video URL views
from flask import (Blueprint, Response, render_template,
                   send_file, make_response)
import json
import os
import string

from config import config
import khan

blueprint = Blueprint('video_views', __name__,
                      template_folder='templates', static_folder='khanvideo')


tree_cache = None


def get_tree():
    global tree_cache
    #khan_webm_dir = config().get_path('VIDEO', 'khan_webm_dir')
    #khan_cache_file = config().get_path('VIDEO', 'khan_cache_file')
    if tree_cache is None:
        tree_cache = get_tree_from_filesystem()
    return tree_cache


def get_tree_from_filesystem():
    khan_webm_dir = config().get_path('VIDEO', 'khan_webm_dir')
    return khan.find_khan(khan_webm_dir)


def split_khanpath(khanpath):
    return [int(x) for x in khanpath.split('/') if x != '']


def join_khanpath(khanpath):
    return string.join([str(x) for x in khanpath], '/')


@blueprint.route('/khanjson/')
@blueprint.route('/khanjson/<path:khanpath>/')
def khan_json_view(khanpath=''):
    path = split_khanpath(khanpath)
    tree = get_tree()
    name, subtree = khan.get(tree, path)
    r = khan.getchildren(tree, path)
    return Response(json.dumps(r, indent=4), mimetype="application/json")


@blueprint.route('/')
@blueprint.route('/khan/')
@blueprint.route('/khan/<path:khanpath>/')
def khan_view(khanpath=''):
    path = split_khanpath(khanpath)
    tree = get_tree()
    name, subtree = khan.get(tree, path)
    r = khan.getchildren(tree, path)
    if 'children' in r:
        def childmap(child):
            return {
                'index': child[0],
                'name': child[1],
                'url': "/iiab/video/khan/" + join_khanpath(path + [child[0]])
            }
        children = map(childmap, r['children'])
        return render_template('khan_index.html', breadcrumbs=r['breadcrumbs'], children=children)
    elif 'file' in r:
        base_url = config().get('VIDEO', 'video_url')
        webm = base_url + '/' + khanpath + ".webm"
        h264 = base_url + '/' + khanpath + ".m4v"
        title = r['breadcrumbs'][-1][1]
        return render_template('khan_video.html', breadcrumbs=r['breadcrumbs'],
                               webm=webm, h264=h264, title=title)
    else:
        raise Exception("Unknown return type in Khan Academy tree")


@blueprint.route('/khanvideo/<path:khanpath>.webm')
def khan_webm_view(khanpath=''):
    print "webm_view", khanpath
    path = split_khanpath(khanpath)
    tree = get_tree()
    filename = khan.getfile(tree, path)
    khan_webm_dir = config().get_path('VIDEO', 'khan_webm_dir')
    r = send_file(os.path.join(khan_webm_dir, filename))
    return make_response(r, 200, {'Accept-Ranges': 'bytes'})


@blueprint.route('/khanvideo/<path:khanpath>.m4v')
def khan_h264_view(khanpath=''):
    path = split_khanpath(khanpath)
    tree = get_tree()
    filename = khan.getfile(tree, path)
    filename = os.path.splitext(filename)[0] + '.m4v'
    khan_webm_dir = config().get_path('VIDEO', 'khan_h264_dir')
    r = send_file(os.path.join(khan_webm_dir, filename), mimetype='video/mp4')
    return make_response(r, 200, {'Accept-Ranges': 'bytes'})



#!/usr/bin/env python

from xml.etree import ElementTree as etree
# lxml is 10 times faster than ElementTree
#from lxml import etree
from base64 import b64decode
import iso639
import os


def getHumanReadableBookId(path):
    """Gets the human readable id for a book from the book's path.
    Uses same technique as kiwix."""
    hid = os.path.basename(path)
    hid = os.path.splitext(hid)[0]
    hid = hid.replace(" ", "_")
    hid = hid.replace("+", "plus")
    return hid


def intWithCommas(x):
    """Pretty print an integer with commas"""
    if type(x) not in [type(0), type(0L)]:
        raise TypeError("Parameter must be an integer.")
    if x < 0:
        return '-' + intWithCommas(-x)
    result = ''
    while x >= 1000:
        x, r = divmod(x, 1000)
        result = ",%03d%s" % (r, result)
    return "%d%s" % (x, result)


def clean_book(book):
    """Fixes up the book data"""
    book2 = {}
    for k, v in book.items():
        if k == "favicon":
            v = b64decode(v)
        elif k == "articleCount":
            v = int(v)
            #book2['articleCountString'] = "{:,}".format(v)  # 2.7
            book2['articleCountString'] = intWithCommas(v)
        elif k == "mediaCount":
            v = int(v)
        elif k == "size":
            v = int(v)
        elif k == "language":
            if v == 'en':  # Mislabel, replace with 3-letter label
                v = 'eng'
            if v in iso639.iso6392:
                book2["languageEnglish"] = iso639.iso6392[v]['english']
            else:
                book2["languageEnglish"] = v
        book2[k] = v
    if 'language' not in book2:
        title = book2.get('title', '')
        if title.find(" ml ") != -1:
            lang = 'mal'
        elif title.find(" zh ") != -1:
            lang = 'zho'
        else:
            # Assume english
            lang = 'eng'
        book2['language'] = lang
        book2['languageEnglish'] = iso639.iso6392[lang]['english']
    hid = getHumanReadableBookId(book2['path'])
    book2['humanReadableId'] = hid
    return book2


def parse_library(library_xml_filename):
    """Parse a kiwix library xml file"""
    with open(library_xml_filename, "r") as f:
        et = etree.parse(f)
        root = et.getroot()
        books = root.findall("book")
    books = map(clean_book, books)
    return books


def get_languages(library):
    """Get a list of all unique languages found in the library,
    sorted in decreasing order of total number of articles in that language"""
    langs = dict()
    for book in library:
        lang = book['language']
        langEng = book['languageEnglish']
        articles = book['articleCount']
        books = []
        if lang in langs:
            articles += langs[lang].get('articleCount', 0)
            books = langs[lang].get('books', [])
        books.append(book)
        langs[lang] = {
            'language': lang,
            'languageEnglish': langEng,
            'articleCount': articles,
            'books': books
        }
    langs = langs.values()
    langs.sort(key=lambda x: -x['articleCount'])
    return langs



# adapted from whoosh.spelling

"""This module contains helper functions for correcting typos in user queries.
"""

from whoosh import query
from whoosh.spelling import QueryCorrector, Correction


# QueryCorrector objects

class MultiFieldQueryCorrector(QueryCorrector):
    """A simple query corrector based on a mapping of field names to
    :class:`Corrector` objects, and a list of ``("fieldname", "text")`` tuples
    to correct. And terms in the query that appear in list of term tuples are
    corrected using the appropriate corrector.
    """

    def __init__(self, correctors, terms, prefix=0, maxdist=2):
        """
        :param correctors: a dictionary mapping field names to
            :class:`Corrector` objects.
        :param terms: a sequence of ``("fieldname", "text")`` tuples
            representing terms to be corrected.
        :param prefix: suggested replacement words must share this number of
            initial characters with the original word. Increasing this even to
            just ``1`` can dramatically speed up suggestions, and may be
            justifiable since spellling mistakes rarely involve the first
            letter of a word.
        :param maxdist: the maximum number of "edits" (insertions, deletions,
            subsitutions, or transpositions of letters) allowed between the
            original word and any suggestion. Values higher than ``2`` may be
            slow.
        """

        self.correctors = correctors
        self.termset = frozenset(terms)
        self.prefix = prefix
        self.maxdist = maxdist

    def correct_query(self, q, qstring):
        correctors = self.correctors
        termset = self.termset
        prefix = self.prefix
        maxdist = self.maxdist

        corrected_tokens = []
        corrected_q = q
        field_names = [ t.fieldname for t in q.all_tokens() ]
        corrections_data = {}

        #    .... maybe see if q.... can filter by field_name.  create multiple correction obj one per fieldname and return list
        for token in q.all_tokens():
            fname = token.fieldname
            if (fname, token.text) in termset:
                sugs = correctors[fname].suggest(token.text, prefix=prefix,
                                                 maxdist=maxdist)
                if sugs:
                    if fname not in corrections_data:
                        corrections_data[fname] = { "corrected_q" : q, "corrected_tokens" : [] }
                    sug = sugs[0]
                    corrections_data[fname]['corrected_q'] = corrections_data[fname]['corrected_q'].replace(token.fieldname, token.text, sug)
                    token.text = sug
                    corrections_data[fname]['corrected_tokens'].append(token)

        return [Correction(q, qstring, corrections_data[f]['corrected_q'], corrections_data[f]['corrected_tokens']) for f in corrections_data]
        #return Correction(q, qstring, corrected_q, corrected_tokens)




# Top level URL views
from flask import Blueprint, make_response, render_template, send_file, send_from_directory

from config import config

blueprint = Blueprint('top_views', __name__,
                      template_folder='templates', static_folder='static')


@blueprint.route('/')
def index():
    error = None
    if config().get_knowledge_dir() is None:
        error = "Could not find knowledge directory containing Internet-in-a-Box dataset.  "
        error += "The configured knowledge_dir path is " + config().get('DEFAULT', 'knowledge_dir')
        error += " and search_for_knowledge_dir is "
        if config().get('DEFAULT', 'search_for_knowledge_dir'):
            error += "ON, so all mounted filesystems were checked."
        else:
            error += "OFF, so other mounted filesystems were NOT checked."
    return render_template("index.html", kiwix_url=config().get('KIWIX', 'url'), error=error)


# This is a hack because of the double //static path issue
#@blueprint.route('static/<path:filename>')
#def static(filename):
#    return blueprint.send_static_file(filename)


@blueprint.route('test')
def test():
    print "TEST"
    return make_response((send_file('/var/www/foo.webm', mimetype="video/webm"), 200, {'Accept-Ranges': 'bytes'}))



# Misc utility functions
# By Braddock Gaskill, Feb 2013
from subprocess import Popen, PIPE
import re


def whoosh2dict(hits):
    """Convert from whoosh results list to
    a list of dictionaries with a key/value pair for
    each schema column"""
    m = []
    for hit in hits:
        d = {}
        for k, v in hit.items():
            d[k] = v
        m.append(d)
    return m


def run_mount():
    """Run the mount command and return the parsed results"""
    p = Popen(['mount'], stdout=PIPE)
    data = p.stdout.readlines()
    regex = '(.*) on (.*) type ([^ ]*) (.*)\n'
    r = []
    for line in data:
        m = re.match(regex, line)
        if m is not None:
            r.append(m.groups())
    return r



import struct
import string
import liblzma
import timepro

HEADER_FORMAT = [
    ('I', 'magicNumber'),
    ('I', 'version'),
    ('Q', 'uuidLow'),
    ('Q', 'uuidHigh'),
    ('I', 'articleCount'),
    ('I', 'clusterCount'),
    ('Q', 'urlPtrPos'),
    ('Q', 'titlePtrPos'),
    ('Q', 'clusterPtrPos'),
    ('Q', 'mimeListPos'),
    ('I', 'mainPage'),
    ('I', 'layoutPage'),
    ('Q', 'checksumPos')
]


ARTICLE_ENTRY_FORMAT = [
    ('H', 'mimetype'),
    ('B', 'parameterLen'),
    ('c', 'namespace'),
    ('I', 'revision'),
    ('I', 'clusterNumber'),
    ('I', 'blobNumber')
    # Null terminated url
    # Null terminated title
    # variable length parameter data
]


REDIRECT_ENTRY_FORMAT = [
    ('H', 'mimetype'),
    ('B', 'parameterLen'),
    ('c', 'namespace'),
    ('I', 'revision'),
    ('I', 'redirectIndex')
    # Null terminated url
    # Null terminated title
    # variable length parameter data
]


CLUSTER_FORMAT = [
    ('B', 'compressionType')
]


# A null byte
NULL = struct.pack('B', 0)


def format_from_rich(rich_format):
    return "<" + string.join([x[0] for x in rich_format], "")


def read_null_terminated(f, encoding='utf-8'):
    s = ""
    while True:
        b = f.read(1)
        if b == NULL:
            return s.decode(encoding)
        s = s + b


def binary_search(f, t, min, max):
    while 1:
        if max < min:
            return None
        m = (min + max) / 2
        v = f(m)
        if v < t:
            min = m + 1
        elif v > t:
            max = m - 1
        else:
            return m


def full_url(namespace, url):
    return namespace + '/' + url


class Format(object):
    def __init__(self, rich_format):
        self.rich_fmt = rich_format
        self.fmt = format_from_rich(rich_format)
        self.compiled = struct.Struct(self.fmt)
        self.size = self.compiled.size

    def unpack_format(self, buffer, offset=0):
        fields = self.compiled.unpack_from(buffer, offset)
        d = []
        for field, entry in zip(fields, self.rich_fmt):
            d.append((entry[1], field))
        return d

    def unpack_format_from_file(self, f, seek=None):
        if seek is not None:
            f.seek(seek)
        buf = f.read(self.size)
        d = self.unpack_format(buf)
        return d

    @timepro.profile()
    def unpack(self, buffer, offset=0):
        """Override this to get more complex behavior"""
        return self.unpack_format(buffer, offset)

    @timepro.profile()
    def unpack_from_file(self, f, seek=None):
        """Override this to get more complex behavior"""
        return self.unpack_format_from_file(f, seek)


class HeaderFormat(Format):
    def __init__(self):
        super(HeaderFormat, self).__init__(HEADER_FORMAT)


class ClusterFormat(Format):
    def __init__(self):
        super(ClusterFormat, self).__init__(CLUSTER_FORMAT)


class ArticleEntryFormat(Format):
    def __init__(self):
        super(ArticleEntryFormat, self).__init__(ARTICLE_ENTRY_FORMAT)

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        d = super(ArticleEntryFormat, self).unpack_from_file(f, seek)
        url = read_null_terminated(f)
        title = read_null_terminated(f)
        parameter = f.read(dict(d)['parameterLen'])
        d.extend([('url', url),
                  ('title', title),
                  ('parameter', parameter)]
                 )
        return d


class RedirectEntryFormat(Format):
    def __init__(self):
        super(RedirectEntryFormat, self).__init__(REDIRECT_ENTRY_FORMAT)

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        d = super(RedirectEntryFormat, self).unpack_from_file(f, seek)
        url = read_null_terminated(f)
        title = read_null_terminated(f)
        parameter = f.read(dict(d)['parameterLen'])
        d.extend([('url', url),
                  ('title', title),
                  ('parameter', parameter)]
                 )
        return d


class MimeTypeListFormat(Format):
    def __init__(self):
        super(MimeTypeListFormat, self).__init__("")

    def unpack(self, buffer, offset=0):
        raise Exception("Unimplemented")

    def unpack_from_file(self, f, seek=None):
        if seek is not None:
            f.seek(seek)
        mimetypes = []
        while True:
            s = read_null_terminated(f)
            if s == "":
                return mimetypes
            mimetypes.append(s)


class ZimFile(object):
    def __init__(self, filename):
        self.filename = filename
        self.redirectEntryFormat = RedirectEntryFormat()
        self.articleEntryFormat = ArticleEntryFormat()
        self.clusterFormat = ClusterFormat()
        self.f = open(filename, "r")
        self.header = dict(HeaderFormat().unpack_from_file(self.f))
        self.mimeTypeList = MimeTypeListFormat().unpack_from_file(self.f, self.header['mimeListPos'])

    @timepro.profile()
    def read_directory_entry(self, offset):
        """May return either a Redirect or Article entry depending on flag"""
        self.f.seek(offset)
        buf = self.f.read(2)
        fields = struct.unpack('H', buf)
        if fields[0] == 0xffff:  # Then redirect
            return dict(self.redirectEntryFormat.unpack_from_file(self.f, offset))
        else:
            return dict(self.articleEntryFormat.unpack_from_file(self.f, offset))

    @timepro.profile()
    def read_url_pointer(self, index):
        self.f.seek(self.header['urlPtrPos'] + 8 * index)
        buf = self.f.read(8)
        fields = struct.unpack('Q', buf)
        return fields[0]

    @timepro.profile()
    def read_cluster_pointer(self, index):
        """Returns a pointer to the cluster and the cluster size"""
        # FIXME: we assume clusters are listed in order to obtain size
        self.f.seek(self.header['clusterPtrPos'] + 8 * index)
        buf = self.f.read(16)
        fields = struct.unpack('QQ', buf)
        # FIXME: If last cluster, we use the checksumPos which is guaranteed to
        # point to 16 bytes before the end of file.
        if index == self.header['clusterCount'] - 1:
            return fields[0], self.header['checksumPos'] - fields[0]
        return fields[0], fields[1] - fields[0]

    @timepro.profile()
    def read_directory_entry_by_index(self, index):
        ptr = self.read_url_pointer(index)
        return self.read_directory_entry(ptr)

    def unpack_blob_ptr(self, buf, index):
        ptr = struct.unpack_from('I', buf, 4 * index)[0]
        return ptr

    @timepro.profile()
    def read_blob(self, cluster_index, blob_index):
        ptr, size = self.read_cluster_pointer(cluster_index)
        cluster = self.clusterFormat.unpack_from_file(self.f, ptr)
        cluster = dict(cluster)
        self.f.seek(ptr + 1)
        timepro.start('read')
        raw = self.f.read(size - 1)
        timepro.end('read')
        if cluster['compressionType'] in [0, 1]:
            pass  # uncompressed
        elif cluster['compressionType'] == 4:
            timepro.start('decompress')
            raw = liblzma.decompress(raw)
            timepro.end('decompress')
        else:
            raise Exception("Unknown ZIM file compression type in cluster " + str(cluster_index) + " got type " + str(cluster['compressionType']))
        blob0 = self.unpack_blob_ptr(raw, 0)
        # Number of blobs in cluster
        nblob = blob0 / 4
        if nblob < blob_index + 1:
            raise Exception("Blob index specified beyond end of cluster.  Blob " + str(blob_index) + " requested but only " + str(nblob) + " exist")
        p = self.unpack_blob_ptr(raw, blob_index)
        n = self.unpack_blob_ptr(raw, blob_index + 1)  # Next blob
        if n > len(raw):
            raise Exception("Blob specified beyond end of cluster. " + str(p) + " to " + str(n - 1) + " in cluster of length " + str(len(raw)))
        #length = n - p
        return raw[p:n - 1]

    @timepro.profile()
    def get_article_by_index(self, index, follow_redirect=True):
        entry = self.read_directory_entry_by_index(index)
        if 'redirectIndex' in entry.keys():
            if follow_redirect:
                print "REDIRECT TO " + str(entry['redirectIndex'])
                return self.get_article_by_index(entry['redirectIndex'], follow_redirect)
            else:
                return None, entry['redirectIndex'], entry['namespace']
        data = self.read_blob(entry['clusterNumber'], entry['blobNumber'])
        mime = self.mimeTypeList[entry['mimetype']]
        namespace = entry['namespace']
        return data, mime, namespace

    @timepro.profile()
    def get_entry_by_url_linear(self, namespace, url):
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            if entry['url'] == url and entry['namespace'] == namespace:
                return i
        return None

    @timepro.profile()
    def get_entry_by_url(self, namespace, url):
        nsurl = full_url(namespace, url)

        def check(idx):
            entry = self.read_directory_entry_by_index(idx)
            return full_url(entry['namespace'], entry['url'])

        m = binary_search(check, nsurl, 0, self.header['articleCount'])
        if m is None:
            return None
        entry = self.read_directory_entry_by_index(m)
        return entry, m

    def get_article_by_url(self, namespace, url, follow_redirect=True):
        entry, idx = self.get_entry_by_url(namespace, url)
        if idx is None:
            return None
        return self.get_article_by_index(idx, follow_redirect=follow_redirect)

    def get_main_page(self):
        main_index = self.header['mainPage']
        return self.get_article_by_index(main_index)

    def validate(self):
        """This is a mostly a self-test, but will validate various assumptions"""
        # Test that URLs are properly ordered
        last = None
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            assert entry is not None
            nsurl = full_url(entry['namespace'], entry['url'])
            if last is not None:
                assert nsurl > last
            last = nsurl
        timepro.log_all()
        timepro.reset()

        # Test load by url performance
        for i in range(0, self.header['articleCount'], 100):
            entry = self.read_directory_entry_by_index(i)
            entry2, idx = self.get_entry_by_url(entry['namespace'], entry['url'])
            assert entry2 is not None
        timepro.log_all()
        timepro.reset()

        # Test load of the last article
        article, mime, ns = self.get_article_by_index(self.header['articleCount'] - 1)
        entry = self.read_directory_entry_by_index(self.header['articleCount'] - 1)
        entry2, idx = self.get_entry_by_url(entry['namespace'], entry['url'])
        assert entry2 is not None

        # Test load subset of all articles
        for i in range(0, self.header['articleCount'], 100):
            if i % 1000 == 0:
                print i
            article, mime, ns = self.get_article_by_index(i)
            if article is None:  # Redirect
                assert mime is not None
        timepro.log_all()
        timepro.reset()

    def list_articles_by_url(self):
        """Mostly for testing"""
        s = ""
        for i in range(self.header['articleCount']):
            entry = self.read_directory_entry_by_index(i)
            s += full_url(entry['namespace'], entry['url']) + "\n"
        return s



"""Routines for retrieving and converting OpenStreetMap
mod_tile imagery tiles.  By Braddock Gaskill, March 2013"""
import struct
import os
import thread
from md5 import md5
try:
    import progressbar
except ImportError:
    pass


META_MAGIC = "META"
MAX_ZOOM = 18


class TileSet(object):
    def __init__(self, tile_path, xmlname, METATILE=8, flatter=False):
        """flatter=True will turn on the flatter directory paths"""
        self.tile_path = tile_path
        self.xmlname = xmlname
        self.METATILE = METATILE
        self.flatter = flatter

    def xyz_to_meta_deep(self, x, y, z):
        """Deep, sparse directory structure normally used by mod_tile"""
        mask = self.METATILE - 1
        x &= ~mask
        y &= ~mask
        hashes = {}

        for i in range(0, 5):
            hashes[i] = ((x & 0x0f) << 4) | (y & 0x0f)
            x >>= 4
            y >>= 4

        meta = "%s/%s/%d/%u/%u/%u/%u/%u.meta" % (self.tile_path, self.xmlname, z, hashes[4], hashes[3], hashes[2], hashes[1], hashes[0])
        return meta

    def xyz_to_meta_flatter(self, x, y, z):
        """New flatter, denser directory structure"""
        mask = self.METATILE - 1
        x &= ~mask
        y &= ~mask
        hashes = {}

        hashes[0] = ((x & 0xff) << 8) | (y & 0xff)
        x >>= 8
        y >>= 8

        for i in [1, 2, 3]:
            hashes[i] = ((x & 0x0f) << 4) | (y & 0x0f)
            x >>= 4
            y >>= 4

        meta = "%s/%s/%d/%u/%u/%u/%u.meta" % (self.tile_path, self.xmlname, z, hashes[3], hashes[2], hashes[1], hashes[0])
        return meta

    def xyz_to_meta(self, x, y, z):
        if self.flatter:
            return self.xyz_to_meta_flatter(x, y, z)
        return self.xyz_to_meta_deep(x, y, z)

    def xyz_to_meta_offset(self, x, y, z):
        mask = self.METATILE - 1
        offset = (x & mask) * self.METATILE + (y & mask)
        return offset


class TileNotFoundException(Exception):
    pass


class TileInvalidFormat(Exception):
    pass


def progress_bar(name, maxval):
    widgets = [name, progressbar.Percentage(), ' ', progressbar.Bar(), ' ', progressbar.ETA()]
    pbar = progressbar.ProgressBar(widgets=widgets, maxval=maxval)
    pbar.start()
    return pbar


def meta_write(tileset, x, y, z, indices, blobs):
    """Writes a meta tile file consisting of a series of blobs.  indicies is a
    list of blob indexes, or -1 if index entry has no blob"""
    METATILE = tileset.METATILE

    assert(len(indices) == METATILE * METATILE)

    # Get path and make parent dirs
    meta_path = tileset.xyz_to_meta(x, y, z)
    d = os.path.dirname(meta_path)
    if not os.path.exists(d):
        try:
            os.makedirs(d)
        except OSError:
            # Multiple threads can race when creating directories,
            # ignore exception if the directory now exists
            if not os.path.exists(d):
                raise

    # Assert that every blob is referenced by an index
    blob_indices = set(range(len(blobs)))
    used_indices = set([x for x in indices if x != -1])
    assert(len(blob_indices - used_indices) == 0)

    # Header size
    offset = len(META_MAGIC) + 4 * 4
    # Need to pre-compensate the offsets for the size of the offset/size table we are about to write
    offset += (2 * 4) * (METATILE * METATILE)

    # Compute offsets for blobs
    bloboffsets = [0] * len(blobs)
    for i in range(len(blobs)):
        bloboffsets[i] = offset
        offset += len(blobs[i])

    # Calculate the offsets and sizes tables
    offsets = [0] * (METATILE * METATILE)
    sizes = [0] * (METATILE * METATILE)

    for i in range(len(indices)):
        bid = indices[i]
        if bid != -1:
            offsets[i] = bloboffsets[bid]
            sizes[i] = len(blobs[bid])

    # Create a temp file so our save is atomic
    tmp = "%s.tmp.%d" % (meta_path, thread.get_ident())
    f = open(tmp, "w")

    # Header
    f.write(struct.pack("4s4i", META_MAGIC, METATILE * METATILE, x, y, z))

    # Write out the offset/size table
    for mt in range(0, METATILE * METATILE):
        f.write(struct.pack("2i", offsets[mt], sizes[mt]))

    # Write out the blobs
    for i in range(len(blobs)):
        f.write(blobs[i])

    # Close and atomically rename file
    f.close()
    os.rename(tmp, meta_path)


def meta_save(tileset, x, y, z, tiles):
    """tiles is a list of (x, y, bytes).
    All tiles must be within a single meta file specified
    by tileset/x/y/z.  tiles are stored on disk in the order in which they are
    listed"""
    METATILE = tileset.METATILE
    meta_filename = tileset.xyz_to_meta(x, y, z)

    # Check that all tiles are in the same meta file
    tile_xy = set()
    for (tile_x, tile_y, tile) in tiles:
        tile_meta = tileset.xyz_to_meta(tile_x, tile_y, z)
        if tile_meta != meta_filename:
            raise Exception("Tile not within meta file")
        if (tile_x, tile_y) in tile_xy:
            raise Exception("Multiple tiles with same x,y coordinates")
        tile_xy.add((tile_x, tile_y))

    # Calculate md5sums for all tiles
    hashes = [0] * len(tiles)
    for i, (tile_x, tile_y, tile) in enumerate(tiles):
        hashes[i] = md5(tile).digest()

    # Calculate indices for each tile
    indices = [-1] * (METATILE * METATILE)
    for i, (tile_x, tile_y, tile) in enumerate(tiles):
        index = tileset.xyz_to_meta_offset(tile_x, tile_y, z)
        indices[index] = hashes[i]

    unique_hashes = {}
    blob_index = 0
    for i, entry in enumerate(tiles):
        if hashes[i] not in unique_hashes:  # First tile of this md5
            unique_hashes[hashes[i]] = blob_index
            blob_index += 1

    # map indices unique hashes to blob index
    for i, entry in enumerate(indices):
        if indices[i] != -1:
            indices[i] = unique_hashes[indices[i]]

    # Create blobs
    blobs = []
    blob_index = 0
    for i, (tile_x, tile_y, tile) in enumerate(tiles):
        tile_hash = hashes[i]
        if unique_hashes[tile_hash] == blob_index:  # First unique tile
            blobs.append(tile)
            blob_index += 1

    assert(blob_index == len(blobs))

    #print "compressed %i to %i for %i indices" % (len(tiles), len(blobs), len([x for x in indices if x != -1]))
    # Save the file
    meta_write(tileset, x, y, z, indices, blobs)


def meta_load_index(tileset, x, y, z):
    """Opens a meta tile file and reads the index.
    returns (file_descriptor, offsets, sizes)"""
    global META_MAGIC
    METATILE = tileset.METATILE
    meta_path = tileset.xyz_to_meta(x, y, z)
    if not os.path.exists(meta_path):
        raise TileNotFoundException("Tile file not found at " + meta_path)
    f = open(meta_path, 'r')
    offset = len(META_MAGIC) + 4 * 4
    offset += (2 * 4) * (METATILE * METATILE)
    magic = f.read(4)
    if magic != META_MAGIC:
        raise TileInvalidFormat("Tile " + meta_path + " is not a valid tile.  Magic " + META_MAGIC + " is not present")
    fmt = "4i"
    (n, x, y, z) = struct.unpack(fmt, f.read(struct.calcsize(fmt)))
    offsets = []
    sizes = []
    for mt in range(0, METATILE * METATILE):
        (offset, size) = struct.unpack('2i', f.read(2 * 4))
        offsets.append(offset)
        sizes.append(size)
    return (f, offsets, sizes)


def meta_load_all(tileset, x, y, z):
    """Read all tiles out of a meta tile"""
    METATILE = tileset.METATILE
    f, offsets, sizes = meta_load_index(tileset, x, y, z)
    tiles = []
    for xx in range(0, METATILE):
        for yy in range(0, METATILE):
            offset = offsets[xx * METATILE + yy]
            size = sizes[xx * METATILE + yy]
            if offset != 0:
                f.seek(offset)
                tile = f.read(size)
                tiles.append((x + xx, y + yy, tile))
    return tiles


def meta_load_one(tileset, x, y, z):
    """Read a single tile out of the appropriate meta tile file"""
    f, offsets, sizes = meta_load_index(tileset, x, y, z)
    index = tileset.xyz_to_meta_offset(x, y, z)
    offset = offsets[index]
    size = sizes[index]
    f.seek(offset)
    tile = f.read(size)
    return tile


def convert(src, dst, z):
    """Convert a source mod_tile tree of one METATILE setting
    to a new tree of a larger (multiple) METATILE setting.
    Example usage:
        import osmtile
        src=osmtile.TileSet('/knowledge/processed/mod_tile', 'default', METATILE=8, flatter=False)
        dst=osmtile.TileSet('/knowledge/modules/openstreetmap/mod_tile64', 'default', METATILE=64, flatter=True)
        for z in range(0, 16):
            osmtile.convert(src, dst, z)
    """
    assert(dst.METATILE > src.METATILE)
    assert(dst.METATILE % src.METATILE == 0)
    size = 2 ** z
    progress = progress_bar("Level " + str(z) + " Tiles ", size)
    for y in xrange(0, size, dst.METATILE):
        progress.update(y)
        for x in xrange(0, size, dst.METATILE):
            tiles = []
            for src_y in xrange(y, y + dst.METATILE, src.METATILE):
                for src_x in xrange(x, x + dst.METATILE, src.METATILE):
                    try:
                        tiles.extend(meta_load_all(src, src_x, src_y, z))
                    except TileNotFoundException:
                        pass
                    except TileInvalidFormat as e:
                        print "ERROR reading tile: " + e.message
            if len(tiles) > 0:
                meta_save(dst, x, y, z, tiles)
    progress.update(size)



from flask import Flask, request
#from flask.ext.mako import MakoTemplates
from flaskext.babel import Babel
from flask.ext.autoindex import AutoIndex

from config import config
import top_views
#import search_views
import map_views
import video_views
import gutenberg
import wikipedia_views
import zim_views
from extensions import db
import sys


def create_app(debug=True, enable_profiler=False, profiler_quiet=False):
    """
    :param debug: whether to configure app for debug
    :param enable_profiler: enable flask profiler, causes function to return ProfilerMiddleware rather than Flask object which can be started by run_simple
    :param profiler_quiet: when profiler enabled sets whether profiler output echoed to stdout
    """
    app = Flask("iiab")
    app.url_map.strict_slashes = False
    app.use_x_sendfile = config().getboolean('WEBAPP', 'use_x_sendfile')

    # Configuration items
    if debug:
        app.debug = True
        app.config['DEBUG'] = True
        app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0

    base_prefix = config().get('WEBAPP', 'base_prefix')

    blueprints = [
        (top_views.blueprint, base_prefix),
        #(search_views.blueprint, base_prefix),
        (gutenberg.gutenberg, base_prefix + "books"),
        (map_views.blueprint, base_prefix + "maps"),
        (video_views.blueprint, base_prefix + "video"),
        (wikipedia_views.blueprint, base_prefix + "wikipedia"),
        (zim_views.blueprint, base_prefix + "zim")
    ]
    for blueprint, prefix in blueprints:
        app.register_blueprint(blueprint, url_prefix=prefix)

    # set global config variables referenced by SQLAlchemy
    app.config['SQLALCHEMY_ECHO'] = config().getboolean('GUTENBERG', 'sqlalchemy_echo')
    app.config['SQLALCHEMY_DATABASE_URI'] = config().get_path('GUTENBERG', 'sqlalchemy_database_uri')

    configure_babel(app)
    db.init_app(app)

    if enable_profiler:
        from werkzeug.contrib.profiler import ProfilerMiddleware, MergeStream
        f = open('profiler.log', 'w')
        if profiler_quiet:
            app = ProfilerMiddleware(app, f)
        else:
            stream = MergeStream(sys.stdout, f)
            app = ProfilerMiddleware(app, stream)

    # Auto Index the software repository
    autoindex = AutoIndex(app, add_url_rules=False)
    # FIXME: this should be done more elegantly -bcg'13

    @app.route(base_prefix + 'software/<path:path>')
    @app.route(base_prefix + 'software')
    def software_view(path='.'):
        software_dir = config().get_path('SOFTWARE', 'software_dir')
        return autoindex.render_autoindex(path, browse_root=software_dir, endpoint='software_view')

    print "URL MAP: ", app.url_map

    return app


def configure_babel(app):
    # flask-babel
    babel = Babel(app)

    @babel.localeselector
    def get_locale():
        accept_languages = config().get_json('GUTENBERG', 'babel_accept_languages')
        return request.accept_languages.best_match(accept_languages)



import re
import os

import kiwix
from zimpy import ZimFile


def replace_paths(top_url, html):
    replace = u"\\1\\2" + top_url + "/\\3/"
    html = re.sub(u'(href|src)(=["\']/)([A-Z\-])/', replace, html)
    html = re.sub(u'(@import[ ]+)(["\']/)([A-Z\-])/', replace, html)
    return html


class Library(object):
    def __init__(self, library_filename):
        self.books = kiwix.parse_library(library_filename)
        self.path = os.path.dirname(library_filename)
        self.readableToBooks = {}
        for book in self.books:
            hid = book['humanReadableId']
            filename = os.path.join(self.path, book['path'])
            self.readableToBooks[hid] = ZimFile(filename)

    def get_article_by_url(self, humanReadableId, namespace, url):
        zimfile = self.get_zimfile(humanReadableId)
        if zimfile is None:
            return None
        article = zimfile.get_article_by_url(namespace, url)
        return article

    def get_zimfile(self, humanReadableId):
        return self.readableToBooks.get(humanReadableId, None)



from setuptools import setup

setup(
    name='Internet-in-a-Box',
    version='0.3.2',
    author="Braddock Gaskill",
    author_email="braddock@braddock.com",
    license="LICENSE.txt",
    long_description="Sharing the world's Free information",
    packages=['iiab'],
    scripts=['iiab-server', 'iiab.wsgi'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Flask',
        'Flask-Babel',
        'Flask-AutoIndex',
        'Flask-SQLAlchemy',
        'whoosh',
        'pyliblzma'
    ]
)



# -*- coding: utf-8 -*-

import os
import re

from flask import (Blueprint, render_template, current_app, request, Response,
                   flash, url_for, redirect, safe_join, make_response,
                   send_file, send_from_directory)
from flaskext.babel import gettext as _
import json

from contextlib import closing

from whoosh.index import open_dir
from whoosh.qparser import MultifieldParser
from .whoosh_multi_field_spelling_correction import MultiFieldQueryCorrector

from .extensions import db
from gutenberg_models import (GutenbergBook, GutenbergFile,
                              GutenbergCreator, gutenberg_books_creator_map)
from config import config

import pagination_helper
from .endpoint_description import EndPointDescription

gutenberg = Blueprint('gutenberg', __name__, url_prefix='/books')
etext_regex = re.compile(r'^etext(\d+)$')
DEFAULT_SEARCH_COLUMNS = ['title', 'creator', 'contributor']  # names correspond to fields in whoosh schema
DEFAULT_RESULTS_PER_PAGE = 20


@gutenberg.route('/')
def index():
    return render_template('gutenberg/index.html')


@gutenberg.route('/search')
def search():
    query = request.args.get('q', '').strip()
    pagination = None
    if query:
        page = int(request.args.get('page', 1))
        (pagination, suggestion) = paginated_search(query, page)
    else:
        flash(_('Please input keyword(s)'), 'error')
    #print pagination.items
    return render_template('gutenberg/search.html', pagination=pagination, keywords=query, suggestion=suggestion, fn_author_to_query=author_to_query, endpoint_desc=EndPointDescription('gutenberg.search', None))


@gutenberg.route('/mirror/<path:filename>')
def gutenberg_mirror(filename):
    mirror_dir = config().get_path('GUTENBERG', 'gutenberg_mirror')
    print "mirror", mirror_dir, filename
    r = send_from_directory(mirror_dir, filename)
    return make_response(r, 200, {'Accept-Ranges': 'bytes'})


def author_to_query(author):
    """Helper function for template macro to convert an author string into a
    search query.

    :param author: unicode string from creator or contributor field of gutenberg index
    :returns: whoosh query string to search creator/contributor fields for given author.
    """
    # contributor field provides extra details about contirbutor's role in brackets -- strip that off so we can search for author in any role.
    author = re.sub(r'\[[^\]]+\]', '', author).strip()
    return u'creator:"{0}" OR contributor:"{0}"'.format(author)


def paginated_search(query_text, page=1, pagelen=DEFAULT_RESULTS_PER_PAGE):
    """
    Return a tuple consisting of an object that emulates an SQLAlchemy pagination object and corrected query suggestion
    pagelen specifies number of hits per page
    page specifies page of results (first page is 1)
    """
    index_dir = config().get_path('GUTENBERG', 'index_dir')
    query_text = unicode(query_text)  # Must be unicode
    ix = open_dir(index_dir)
    sort_column = 'creator'
    with ix.searcher() as searcher:
        query = MultifieldParser(DEFAULT_SEARCH_COLUMNS, ix.schema).parse(query_text)
        try:
            # search_page returns whoosh.searching.ResultsPage
            results = searcher.search_page(query, page, pagelen=pagelen, sortedby=sort_column)
            total = results.total
        except ValueError:  # Invalid page number
            results = []
            total = 0
        paginate = pagination_helper.Pagination(page, pagelen, total, [dict(r.items()) for r in results])
        corrections = deduplicate_corrections(get_query_corrections(searcher, query, query_text))  # list of Corrector objects

        #hf = whoosh.highlight.HtmlFormatter(classname="change")
        #html = corrections.format_string(hf)
        return (paginate, [c.string for c in corrections])


def deduplicate_corrections(corrections):
    """
    Return list of correction that omits entries where the query is unmodified
    :param corrections: list of Corrector objects
    :returns: list of Corrector objects
    """
    # Using values from a dictionary comprehension rather than a list comprehension in order to deduplicate
    #return {c.string : c for c in corrections if c.original_query != c.query}.values()
    # We can't use dictionary comprehension because we are stuck on python 2.6 for Debian stable
    return dict((c.string, c) for c in corrections if c.original_query != c.query).values()


def get_query_corrections(searcher, query, qstring):
    """
    Suggest alternate spelling for search terms by searching each column with
    spelling correction support in turn.

    :param searcher: whoosh searcher object
    :param query: whoosh query object
    :param qstring: search string that was passed to the query object
    :returns: MultiFieldQueryCorrector with one corrector for each corrected column
    """
    fieldnames = [name for name, field in searcher.schema.items() if field.spelling]
    correctors = {}
    for fieldname in fieldnames:
        if fieldname not in correctors:
            correctors[fieldname] = searcher.corrector(fieldname)
    terms = []
    for token in query.all_tokens():
        if token.fieldname in correctors:
            terms.append((token.fieldname, token.text))

    return MultiFieldQueryCorrector(correctors, terms, prefix=2, maxdist=1).correct_query(query, qstring)


@gutenberg.route('/titles')
def by_title():
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergBook.query.order_by(GutenbergBook.title_order).paginate(page, per_page)
    return render_template('gutenberg/title-index.html', pagination=pagination, fn_author_to_query=author_to_query, endpoint_desc=EndPointDescription('.by_title', dict(per_page=per_page)))


@gutenberg.route('/authors')
def by_author():
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergCreator.query.order_by(GutenbergCreator.creator).paginate(page, per_page)

    return render_template('gutenberg/author-index.html', pagination=pagination, endpoint_desc=EndPointDescription('.by_author', dict(per_page=per_page)))


@gutenberg.route('/author/<authorId>')
def author(authorId):
    page = int(request.args.get('page', 1))
    per_page = int(request.args.get('per_page', DEFAULT_RESULTS_PER_PAGE))
    pagination = GutenbergBook.query.filter(gutenberg_books_creator_map.c.creator_id == authorId).filter(gutenberg_books_creator_map.c.book_id == GutenbergBook.textId).paginate(page, per_page)
    return render_template('gutenberg/title-index.html', pagination=pagination, fn_author_to_query=author_to_query, endpoint_desc=EndPointDescription('.author', dict(authorId=authorId, per_page=per_page)))


def mirror_path(filename):
    """Determines the full path to a gutenberg text in the mirror"""
    mirror_dir = config().get_path('GUTENBERG', 'gutenberg_mirror')
    path = os.path.join(mirror_dir, filename)
    return path


def mirror_exists(file_rec):
    path = mirror_path(file_rec.file)
    return os.path.exists(path)


@gutenberg.route('/text/<textId>/details')
def text(textId):
    # Profiling results showing occasional lags.
    # Lags can be minimized by disabling SQLALCHEMY_ECHO and
    # and debug mode.
    # Tested no options, joinedload and subqueryload with no consistent winner.
    record = GutenbergBook.query.filter_by(textId=textId).first()
    # if blueprint has a different static_folder specified we might need to use blueprint.static_folder but currently None
    for x in record.gutenberg_files:
        if not mirror_exists(x):
            print "WARNING: Gutenberg file " + mirror_path(x.file) + " not found"
    record.gutenberg_files = filter(mirror_exists, record.gutenberg_files)

    # fields format is list of tuples:
    # (Table row heading for display, gutenberg_books col name,
    #  sub-table col name if applicable)
    fields = [
        (_('Title'), 'title', ''),
        (_('Author'), 'gutenberg_creators', 'creator'),
        (_('Contributor'), 'gutenberg_contributors', 'contributor'),
        (_('Subject'), 'gutenberg_subjects', 'subject'),
        (_('Category'), 'gutenberg_categories', 'category'),
        (_('Language'), 'gutenberg_languages', 'language')
    ]
    return render_template('gutenberg/book_details.html', record=record, fields=fields)


@gutenberg.route('/text/<textId>/<int:textIndex>')
def read(textId, textIndex):
    data_dir = config().get_path('GUTENBERG', 'root_dir')
    files = GutenbergFile.query.filter_by(textId=textId).all()
    assert textIndex >= 0 and textIndex < len(files)
    fullpath = safe_join(data_dir, files[textIndex].file)
    return send_file(fullpath)


def choose_file(textId):
    files = GutenbergFile.query.filter_by(textId=textId)
    #for f in files:
    #    print f
    return files[0].file


@gutenberg.route('/autocomplete')
def autocomplete():
    term = request.args.get('term', '')
    if term != '':
        index_dir = config().get_path('GUTENBERG', 'index_dir')
        ix = open_dir(index_dir)
        with ix.searcher() as searcher:
            # might use whoosh.analysis.*Analyzer to break query up
            # for matching. However it isn't clear how to combine completion
            # of partial matches across several different columns without
            # lots of effort

            # Be aware that returning a json top-level array leaves us vulnerable to CSRF.
            # http://flask.pocoo.org/docs/security/ In this case this is not of significant
            # concern because the information is not sensitive.  We use a top-level array
            # because this is what jquery autocomplete demands for use without modification.
            suggestions = get_autocomplete_matches(term)
            return Response(response=json.dumps(suggestions), mimetype="application/json")
    else:
        # Choosing an inefficient redirect because still testing different
        # approaches and its easier to centralize the handling.  If we keep
        # this approach we can just change the referencing url
        return redirect(url_for("static", filename="gutenberg_wordlist.json"))


def get_autocomplete_matches(prefix, limit=10):
    def get_prefix_like(prefix):
        (result, _) = re.subn(r'\\', u'\\\\', prefix)
        (result, _) = re.subn(r'_', u'\_', result)
        (result, _) = re.subn(r'%', u'\%', result)
        (result, _) = re.subn(r'\s+', u'%', result)
        return '%' + result + '%'

    def make_sql(colname, tablename, limit):
        sql = "SELECT {0}, downloads FROM {1} WHERE {0} LIKE :like_clause ESCAPE '\\' ORDER BY downloads LIMIT {2};".format(colname, tablename, limit)
        return sql

    like_clause = get_prefix_like(prefix)
    search_fields = [('title', 'gutenberg_books'),
                     ('creator', 'gutenberg_creators'),
                     ('contributor', 'gutenberg_contributors')]
    results = []
    with closing(db.engine.connect()) as conn:
        for colname, tablename in search_fields:
            results.extend(conn.execute(make_sql(colname, tablename, limit), like_clause=like_clause).fetchall())
    return [row[0] for row in sorted(results, key=lambda r: r[1], reverse=True)]
